/* xn-todos.c:  Pilot todo list conduit
 *
 * Copyright 1996, Robert A. Kaplan
 * Rewritten to handle installing to Palm and reading from it, like
 * xn-addresses.c
 *   Michael J. Hammel, October 2000 (mjhammel@graphics-muse.org)
 *
 * This is free software, licensed under the GNU Public License V2.
 * See the file COPYING for details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <strings.h>
#include "pi-source.h"
#include "pi-socket.h"
#include "pi-dlp.h"
#include "pi-todo.h"
#include "pi-xn.h"

#ifdef DEBUG
#undef DEBUG
#endif

typedef struct _xnotes_cat {
	struct _xnotes_cat	*next;
	char						*name;
} XNotesCat_T;

XNotesCat_T	*xnotes_categories = NULL;

char magicmark[0xffff];

char		*progname;

#define CATSXML		"<XNCat>"
#define CATEXML		"</XNCat>"

/*========================================================================
 *	Name:			Help
 *	Prototype:	void Help()
 *					
 *	Description:
 * Print out a help message.
 *
 *	Notes:
 *========================================================================*/
void 
Help(void)
{
  fprintf(stderr,
    "usage:%s [-q] [-p port] [-d category] -r|-w file\n",
    progname);
  fprintf(stderr, "  -q          = do not prompt for HotSync button press\n");
  fprintf(stderr, "  -p port     = use device file <port> to communicate with Pilot\n");
  fprintf(stderr, "  -d category = DELETE old Pilot records in <category>\n");
  fprintf(stderr, "  -r file     = replace existing Todo DB with records read from <file>\n");
  fprintf(stderr, "  -w file     = get records from Pilot and write them to <file>\n");
  exit(2);
}

/*========================================================================
 *	Name:			match_category
 *	Prototype:	int match_category(char *category, struct TodoAppInfo *aai)
 *					
 *	Description:
 * Verify that the category provided is valid based on the categories listed in
 * the tai structure.
 *
 * Returns
 * The index id for the category if it exists.
 * -1 if it doesn't.
 *
 *	Notes:
 *========================================================================*/
int 
match_category(
	char							*buf, 
	struct ToDoAppInfo		*tai
)
{
	int i;
	for (i=0;i<16;i++)
	if (strcasecmp(buf, tai->category.name[i])==0)
		return i;

	/* Wasn't found - return 0, which equates to the "Unfiled" category. */
	return 0;
}

/*========================================================================
 *	Name:			get_xnotes_categories
 *	Prototype:	void get_xnotes_categories(int sd, int db, struct TodoAppInfo *aai) 
 *					
 *	Description:
 * Look for XNotes category names in the notes fields and save them
 * in a link list.
 *
 *	Notes:
 * We preprocess the XNotes categories from the Todos so we can write them
 * out along with the Pilot Categories before we write out the actual Todo 
 * entries.
 *========================================================================*/
void 
get_xnotes_categories(
	int						sd,
	int						db,
	struct ToDoAppInfo	*tai
)
{
	int				i;
	struct ToDo		todo;
	int				attribute, category;
	int				len;
	int				found;
	char				buf[0xffff];
	recordid_t		RecordID;
	char				*value;
	char				*ptag, *ps, *pe;
	XNotesCat_T		*entry;
	XNotesCat_T		*last_entry = NULL;
	XNotesCat_T		*current_entry = NULL;

	for (i=0;;i++) 
	{
		len = dlp_ReadRecordByIndex(sd, db, i, buf, 
					&RecordID, 0, &attribute, &category);

		if(len<0)
			break;

		/* Skip deleted records */
		if((attribute & dlpRecAttrDeleted) || (attribute & dlpRecAttrArchived))
			continue;
  		
		unpack_ToDo(&todo, buf, len);

		/* If there is a category tag in the note... */
		if ( (todo.note) && ( (ptag=strstr(todo.note, CATSXML)) != NULL ) )
		{
			/* Pull the value of that tag out of the note */
			ps = index((char *)(ptag), '>');
			pe = index((char *)(ps), '<');
			value = (char *)malloc(pe-ps);
			bzero(value, pe-ps);
			strncpy(value, (char *)(ps+1), pe-ps-1);

			/* If this category doesn't already exist... */
			if ( match_category(value,tai) == 0 )
			{
				/* And is not already in the saved list... */
				found = 0;
				current_entry = xnotes_categories;
				while(current_entry)
				{
					if (strcmp(current_entry->name, value) == 0 )
					{
						found = 1;
						break;
					}
					current_entry = current_entry->next;
				}
				
				if ( !found )
				{
					/* Save it the category. */
					entry = (XNotesCat_T *)malloc(sizeof(XNotesCat_T));
					bzero((char *)entry, sizeof(XNotesCat_T));
					entry->name = (char *)malloc(strlen(value)+1);
					strcpy(entry->name, value);
					if ( xnotes_categories == NULL )
						xnotes_categories = entry;
					else
						last_entry->next = entry;
					last_entry = entry;
				}
			}
		}

		/*
		 * This frees up space allocated in the struct, not the struct
		 * itself.
		 */
		free_ToDo(&todo);
	}
}

/*========================================================================
 *	Name:			write_categories
 *	Prototype:	void write_categories(FILE *out, struct ToDoAppInfo *tai)
 *					
 *	Description:
 * Writes the list of available categories to file.
 *
 * Input arguments:
 * struct	ToDoAppInfo *tai	Pilot Todo db's category info, etc.
 * 
 * Output arguments:
 *	FILE		*out			file descriptor to write records to local file
 *
 *	Notes:
 *========================================================================*/
int 
write_categories(
	FILE							*out, 
	struct ToDoAppInfo		*tai
)
{
	int			i;
	XNotesCat_T	*entry;

	/* Write out the Pilot's categories. */
	for (i=0;i<16;i++) 
	{
		if ( strlen(tai->category.name[i]) > 0 )
			fprintf(out, "%s\n", tai->category.name[i]);
	}

	/* Add any XNotes specific categories we hid in the Notes sections. */
	entry = xnotes_categories;
	while (entry != NULL)
	{
		fprintf(out, "%s\n", entry->name);
		entry = entry->next;
	}

	return 0;
}

/*========================================================================
 *	Name:			write_file
 *	Prototype:	void write_file(FILE *out, int sd, int db, struct TodoAppInfo *aai) 
 *					
 *	Description:
 * Reads Todo records from the pilot and writes them to a file.
 *
 * Input arguments:
 *	int		sd				file descriptor used to communicate with pilot 
 *	int		db				database handle information for Pilot Todo db
 * struct TodoAppInfo *aai		Pilot Todo db's category info, etc.
 * 
 * Output arguments:
 *	FILE		*out			file descriptor to write records to local file
 *
 *	Notes:
 * If an XNotes XML tag is found in the note then it was an unknown
 * category when it was saved to the pilot.  Use this category under XNotes
 * *unless* the tag matches an existing category (the user may have added it
 * to the Pilot).  In any case, clear the tag from the Notes field.
 *========================================================================*/
int 
write_file(
	FILE						*out,
	int						sd,
	int						db,
	struct ToDoAppInfo	*tai
)
{
	int				i;
	struct ToDo		todo;
	int				attribute, category;
	int				len;
	char				buf[0xffff];
	recordid_t		RecordID;
	char				*tmpcat;
	char				*value;
	char				*ptag, *ps, *pe;

	/*
	 * First, clean up the database. This will actually remove the
	 * archived and deleted files on the Pilot so we don't download them.
	 */
	dlp_CleanUpDatabase(sd, db);

	for (i=0;;i++) 
	{
		len = dlp_ReadRecordByIndex(sd, db, i, buf, 
					&RecordID, 0, &attribute, &category);

		if(len<0)
			break;

		/* Skip deleted records */
		if((attribute & dlpRecAttrDeleted) || (attribute & dlpRecAttrArchived))
			continue;
  		
		unpack_ToDo(&todo, buf, len);

		/* 
		 * Preprocess category/note data.
		 */

		/* Save a copy of the category. */
		tmpcat = (char *)malloc(strlen(tai->category.name[category])+1);
		strcpy(tmpcat, tai->category.name[category]);

		/* If there is a category tag in the note... */
		if ( (todo.note) && ( (ptag=strstr(todo.note, CATSXML)) != NULL ) )
		{
			/* Pull the value of that tag out of the note */
			ps = index((char *)(ptag), '>');
			pe = index((char *)(ps), '<');
			value = (char *)malloc(pe-ps);
			bzero(value, pe-ps);
			strncpy(value, (char *)(ps+1), pe-ps-1);

			/* If this category doesn't already exist... */
			if ( match_category(value,tai) == 0 )
			{
				/* Save it as the category. */
				free(tmpcat);
				tmpcat = (char *)malloc(strlen(value)+1);
				strcpy(tmpcat, value);
			}

			/* Remove the tag from the note. */
			bzero((char *)(ptag), 1);
		}

		/* Write out the record to a file. */
		fprintf(out, "%sAttribute: %d\n", PREFIX, attribute);
		fprintf(out, "%sRecordID: %d\n", PREFIX, (int)RecordID);

		/* fprintf(out, "%sCategory: %s\n", PREFIX, tai->category.name[category]);
		 */
		fprintf(out, "%sCategory: %s\n", PREFIX, tmpcat);

		fprintf(out, "%sPriority: %d\n", PREFIX, todo.priority);
		fprintf(out, "%sCompleted: %s\n", PREFIX, todo.complete ? "Yes" : "No");

		if(todo.indefinite)
			fprintf(out, "%sDue: No Date\n", PREFIX);
		else
			fprintf(out, "%sDue: %s", PREFIX, asctime(&todo.due));

		if(todo.description)
			fprintf(out, "%sDescription: %s\n", PREFIX, todo.description);

		if(todo.note)
			fprintf(out, "%sNote: %s\n", PREFIX, todo.note );

		/*
		 * This frees up space allocated in the struct, not the struct
		 * itself.
		 */
		free_ToDo(&todo);
		free(tmpcat);
	}


	return 0;
}


/*========================================================================
 *	Name:			read_file
 *	Prototype:	void read_file(FILE *in, int sd, int db, struct ToDoAppInfo *tai)
 *					
 *	Description:
 * Read in entries from a text file, formatted the same way we write to
 * text files, and upload it to the pilot.
 *
 * Input arguments:
 *	FILE		*out			file descriptor to reads records from
 *	int		sd				file descriptor used to communicate with pilot 
 *	int		db				database handle information for Pilot ToDo db
 * struct ToDoAppInfo *tai		Pilot ToDo db's category info, etc.
 *
 * Output arguments:
 *
 * Returns
 * 0 always
 *
 *	Notes:
 * Categories from XNotes that are not known to the Pilot are saved under
 * Unfiled on the Pilot and the XNotes category name saved in the Note field
 * inside XML tags.  Its clunky, but its KISS for XNotes.
 *========================================================================*/
int
read_file(
	FILE						*fd,
	int						sd,
	int						db,
	struct ToDoAppInfo	*tai
)
{
	struct ToDo					todo;
	char							*catfield = NULL;
	int							cPtr;
	int							endPtr;
	int							count;
	int							category = 0;
	int							entry_number;
	int							filelen;
	char							*file_text;
	char							*name, *value;
	int							j;
	int							RecordID = 0;
	int							flags = 0;
	int							ToDo_size;
	unsigned char				ToDo_buf[0xffff];

#ifdef DEBUG
	int							total = 1;
#endif

	/* Read the file into memory - it's faster this way. */
	fseek(fd, 0, SEEK_END);
	filelen = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	file_text = (char*)malloc(filelen+1);
	if (file_text == NULL) 
	{
		perror("malloc()");
		exit(1);
	}
   fread(file_text, filelen, 1, fd);
	fclose(fd);


	/*
	 * Initialize pointers, counters used to parse input text.
	 */
	cPtr = (int)file_text;
	count = 0;
	entry_number = 1;

	/*
	 * Move past the Magic Mark
	 */
	name=strtok((char *)cPtr, "\n");
	cPtr = cPtr + strlen(name) + 1;

	/*
	 * Skip the category fields from the file first.
	 * When we're done, move past the marker that separates the category
	 * list from the records.
	 */
	while( strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 )
	{
		name=strtok((char *)cPtr, "\n");
		cPtr = cPtr + strlen(name) + 1;
	}

	/* Initialize the structure we'll use for each record. */
	memset(&todo,0,sizeof(todo));

	/* 
	 * Parse the text from the input file.  The format of each entry in
	 * the input files must be in the following name/value formats:
	 *
	 * Category: <single line of data>
	 * RecordID: <single line of data>
	 * Priority: <single line of data>
	 * Completed: <single line of data>
	 * Due: <single line of data>
	 * Description: <multi-line data>
	 * Note: <multi-line data>
	 * <blank line>
	 * 
	 * Name strings are case insensitive.
	 */

	while( cPtr < (int)((int)file_text + filelen) )
	{

		/* Move past the record prefix */
		cPtr = cPtr + strlen(PREFIX);

		/*
		 * Get the current name, value pair.
		 */
		name=strtok((char *)cPtr, ":");
		cPtr = cPtr + strlen(name) + 2;
	
		/* Grab the multiple lines of a "value". */
		value = (char *)cPtr;
		while ( 
			(strncasecmp((char *)cPtr, PREFIX, strlen(PREFIX)) != 0 ) && 
			((int)cPtr < (int)((int)file_text + filelen) )
			)
				cPtr++;
	
		/* Remove trailing newline */
		endPtr = cPtr - 1;
				
		/*
		 * Null terminate the string 
		 */
		strncpy((char *)endPtr, "\0", 1);


		/* Increment the line of the current entry we are on. */
		count++;

		/*
		 * If name = "Note", save the note text.
		 */
		if ( strncasecmp(name, "Note", 4) == 0 )
		{
			if ( catfield ) 
			{
				/* If an XNotes category is in use, stuff it into the note */
				todo.note = 
					(char *)malloc(
							strlen(value)+
							strlen(catfield)+2);
				sprintf(todo.note, "%s\n%s", value, catfield);
				free(catfield);
				catfield = NULL;
			}
			else
			{
				/* Save the note's character string. */
				todo.note = (char *)malloc(strlen(value)+1);
				sprintf(todo.note, "%s", value);
			}
		}

		/*
		 * If name = "Category", make sure it's a known category.
		 */
		if ( strncasecmp(name, "Category", 8) == 0 )
		{

			/*
			 * Is category already in the Pilot's db? 
			 * Can't use match_category() since we need to check for Unfiled too.
			 */
			category = -1;
			for (j = 0; j < 16; j++)
			{
				if (strcasecmp(tai->category.name[j], value) == 0) 
				{
					category = j;
					break;
				}
			}

			/*
			 * If it's not found, stuff into the note field with XML tags.
			 */
			if ( category == -1 )
			{
				/* Set the Pilot's category to Unfiled */
				category = 0;

				/* Add the XN category to the note field (now or later). */
				if ( todo.note )
				{
					/* Allocate a new block large enough for note and category */
					catfield = 
						(char *)malloc(
									strlen(value)+
									strlen(CATSXML)+
									strlen(CATEXML)+
									strlen(todo.note)+2);

					/* Copy in both */
					sprintf(catfield, "%s\n%s%s%s", todo.note, 
						CATSXML, value, CATEXML);

					/* Free old note */
					free(todo.note);

					/* point note to new block */
					todo.note = catfield;
					catfield = NULL;
				}
				else
				{
					/* Allocate a block for the category */
					catfield = 
						(char *)malloc(
							strlen(value)+ strlen(CATSXML)+ strlen(CATEXML)+ 1);

					/* stuff it in there. */
					sprintf(catfield, "%s%s%s", CATSXML, value, CATEXML);
				}

/*
				fprintf(stderr, 
					"%s: category %s not found on PalmPilot\n",
					progname, value);
				fprintf(stderr, "skipping record.\n");
				exit(5);
*/
			}

		}

		/* Convert the RecordID to a numberic value. */
		if ( strncasecmp(name, "RecordID", 8) == 0 )
		{
			RecordID = atoi(value);
		}

		/* Set the Secret flag if this entry is Private. */
		if ( strncasecmp(name, "Attribute", 9) == 0 )
		{
			flags = atoi(value);
		}

		/* Just save the description as is. */
		if ( strncasecmp(name, "Description", 11) == 0 )
		{
			todo.description = (char *)malloc(strlen(value)+1);
			sprintf(todo.description, "%s", value);
		}

		/* Convert the priority to an integer (1-5) */
		if ( strncasecmp(name, "Priority", 8) == 0 )
		{
			todo.priority = atoi(value);
		}

		/* For "Complete", if it's "Yes", it's 1, otherwise it's 0. */
		if ( strncasecmp(name, "Completed", 9) == 0 )
		{
			if ( strncasecmp(name, "Yes", 3) == 0 )
			{
				todo.complete = 1;
			}
			else
			{
				todo.complete = 0;
			}
		}

		/*
		 * Convert the due date to a tm structure - is strptime() POSIX?
		 */
		if ( strncasecmp(name, "Due", 3) == 0 )
		{
			if ( strncasecmp(value, "no date", 7) == 0 )
			{
				/* Indefinite date - use a dummy field */
				todo.indefinite = 1;
				strptime("Sun Jan 1 00:00:00 2000", "%A %b %d %T %Y", 
					&(todo.due));
			}
			else
			{
				todo.indefinite = 0;
				strptime(value, "%A %b %d %T %Y", &(todo.due));
			}
		}


		if ( count == 8 )
		{

#ifdef DEBUG
			/* Debug */
			printf ("Writing to Pilot...%d\n\n", total++);
#endif

			/*
			 * Pack the structure for writing to the Pilot.
			 */
			ToDo_size = pack_ToDo(&todo, ToDo_buf, sizeof(ToDo_buf));

			/*
			 * And finally write it to the Pilot.
			 */
			dlp_WriteRecord(sd, db, 
				flags, 0, category, ToDo_buf, ToDo_size, 0);

			/*
			 * Free up allocated storage.
			 */
			if ( todo.description != (char *)NULL )
				free(todo.description);
			if ( todo.note != (char *)NULL )
				free(todo.note);
			memset(&todo,0,sizeof(todo));
			if ( catfield != (char *)NULL )
			{
				free(catfield);
				catfield = NULL;
			}

			/* Reset the line counter for each entry */
			if ((int)cPtr < (int)((int)file_text + filelen) )
				count = 0;
			else
				break;

#ifdef DEBUG
			/* Debug */
			printf("End entry number: %d\n", entry_number);
#endif
			entry_number++;

		}
	}

   return(0);
}

/*========================================================================
 *	Name:			main
 *					
 *	Description:
 * Let's go!
 *========================================================================*/
int 
main(
	int argc, 
	char *argv[]
)
{
	struct pi_sockaddr		addr;
	struct ToDoAppInfo		tai;
	int							db;
	int							sd;
	struct PilotUser			U;
	int							ret;
	int							length;
	char							buf[0xffff];
	FILE							*fd;
	int							mode = 0;
	int							quiet = 0;
	int							c;
	char							*deletecategory = 0;


	progname = argv[0];

	sprintf(magicmark, MAGICMARK, "todos");

	/*
	 * Check for PILOTPORT environment variable if user doesn't specify the
	 * port to use on the command line.
	 */
	if (getenv("PILOTPORT"))
	{
		strcpy(addr.pi_device,getenv("PILOTPORT"));
	}
	else
	{
		strcpy(addr.pi_device,PILOTPORT);
	}

	/*
	 * If we have less than 3 cmd lin args, print the help message.
	 */
	if (argc < 3)
		Help();


	/*
	 * Parse command line options, if any.
	 */
	while (((c = getopt(argc, argv, "qp:d:rw")) != -1) && (mode == 0))
	{
		switch (c)
		{
		case 'q':
			quiet = 1;
			break;
		case 'p':
			/* optarg is name of port to use instead of $PILOTPORT or /dev/pilot */
			strcpy(addr.pi_device,optarg);
			break;
		case 'd':
			deletecategory = optarg;
			break;
		case 'r':
			mode = 1;
			break;
		case 'w':
			mode = 2;
			break;
		case 'h': case '?':
			Help();
			break;
		}

	}

	/*
	 * -r or -w argument is required.  If not found, print help message.
	 */
	if (mode == 0)
		Help();

	/*
	 * Unless silent run is requested, prompt for pressing the HotSync button on
	 * the Pilot or cradle.
	 */
	if (!quiet)
	{
		fprintf(stderr,
			"Please insert Pilot in cradle on %s and press HotSync button.\n",
			addr.pi_device);
	}


	/*
	 * Get a socket and set up for communicating with the Pilot.
	 */
	if (!(sd = pi_socket(PI_AF_PILOT, PI_SOCK_STREAM, PI_PF_DLP))) 
	{
		perror("pi_socket");
		exit(3);
	}
    
	addr.pi_family = PI_AF_PILOT;
  
	ret = pi_bind(sd, (struct sockaddr*)&addr, sizeof(addr));
	if(ret == -1) 
	{
		perror("pi_bind");
		exit(3);
	}

	ret = pi_listen(sd,1);
	if(ret == -1) 
	{
		perror("pi_listen");
		exit(3);
	}

	sd = pi_accept(sd, 0, 0);
	if(sd == -1) 
	{
		perror("pi_accept");
		exit(3);
	}

	/* Ask the pilot who it is. */
	dlp_ReadUserInfo(sd,&U);
  
	/* Tell user (via Pilot) that we are starting things up */
	dlp_OpenConduit(sd);
  
	/* Open the ToDo Pad's database, store access handle in db variable */
	if(dlp_OpenDB(sd, 0, 0x80|0x40, "ToDoDB", &db) < 0) 
	{
		fprintf(stderr, "Unable to open ToDoDB\n");
		dlp_AddSyncLogEntry(sd, "Unable to open ToDoDB.\n");
		exit(3);
	}

	/*
	 * Get the ToDo Pad's Application block.
	 */
	length = dlp_ReadAppBlock(sd, db, 0, (unsigned char *)buf, 0xffff);
	unpack_ToDoAppInfo(&tai, (unsigned char *)buf, length);


	/*
	 * If the request is to retrieve records from the Pilot and save to a local
	 * file (ie the -w option was used)...
	 */
	if (mode == 2)
	{
		/*
		 * Open the output file.
		 */
		fd = fopen(argv[optind],"w");
		if (fd == NULL)
		{
			sprintf(buf, "%s: %s", argv[0], argv[optind]);
			perror(buf);
			exit(1);
		}
		fprintf(fd, "%s\n", magicmark);

		/*
		 * Read from the pilot and saved to a file.
		 */
		get_xnotes_categories(sd, db, &tai);
		write_categories(fd, &tai);
		write_file(fd, sd, db, &tai);
		
		/*
		 * if requested, delete the category and it's entries.  Only do this if
		 * the category is *not* the Unfiled category.
		 */
		if (deletecategory)
			dlp_DeleteCategory(sd, db, match_category(deletecategory,&tai));
		fclose(fd);
	}
	else
	{

		/*
       * The -r option was specified, so write a text file full of
       * Todo entries to the Pilot.
		 */

		/* ... open the output file. */
		fd = fopen(argv[optind], "r");
		if (fd == NULL) 
		{
			sprintf(buf, "%s: %s", argv[0], argv[optind]);
			perror(buf);
			exit(1);
		}

		/*
		 * Delete all the old records first.
		 */
		dlp_DeleteRecord(sd, db, 1, 0);

		/*
		 * Process the records retrieved.
		 */
		read_file(fd, sd, db, &tai);

		/*
		 * if requested, delete the category and it's entries.  Only do this if
		 * the category is *not* the Unfiled category.
		 */
		if (deletecategory)
			dlp_DeleteCategory(sd, db, match_category(deletecategory,&tai));

	}

	/*
	 * Now clean up the database. This will actually remove the
	 * archived and deleted files on the Pilot.
	 */
	dlp_CleanUpDatabase(sd, db);


	/* Close the database */
	dlp_CloseDB(sd, db);

	/*
	 * Tell the user who it is, with a different PC id.
	 */
	U.lastSyncPC = 0x00010000;
	U.successfulSyncDate = time(NULL);
	U.lastSyncDate = U.successfulSyncDate;
	dlp_WriteUserInfo(sd,&U);

	/*
	 * Log the update in the Pilot's transfer log.
	 */
	dlp_AddSyncLogEntry(sd, "Wrote ToDo to Pilot.\n");
  
	/* 
	 * Historical code - leave it in till someone better acquainted with the
	 * Pilot-Link libraries takes it out.
	 */
	dlp_EndOfSync(sd,0);
	pi_close(sd);
  
	return 0;
}
